/* 
 *  $Id: Channel.java,v 1.31 1997/08/29 21:00:41 jrsantos Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ants;

import java.io.*;
import java.net.*;

import utils.*;

/**
 * Channel used by nodes to exchanged capsule with one another.
 * Might also add a scavenger thread to free long-lived threads
 * before the queue is full but during idle time.
 *
 * @author David Wetherall
 */

public abstract class Channel
  extends ManagedObject
{
  private CodeCache codecache;
  boolean started;

  public ChannelAddress source;
  public Node client;
  public int clientAddress;
  protected Xdr txbuf;
  protected Xdr rxbuf;

  private static byte ANEPVERSION = 1;
  private static short ANTSTYPE = 18;
  private static short SRCOPTTYPE = 1;
  private static short DESTOPTTYPE = 2;
  private static int IPV4SCHEME = 1;
  private static int ANTSVERSION = 0x00010001;

  public Xdr serialize(Capsule cap) {
    short packetLen;

    // source, etc. must be set already
    if (cap.src == 0) {
      Entity.warn("SecurityExcpetion: source not set in serialize for "+ cap);
      throw new SecurityException();
    }
        
    // previous is updated for use by the code loading
    cap.previous = clientAddress;
 
    // resources decremented here just before we leave
    if (--cap.resources <= 0) {
      throw new ResourceLimitException();
    }

    Xdr xdr;
    
    if (cap.xdr != null) {
      xdr = cap.xdr;
      xdr.index = 0;
      // SEARCH FOR SOURCE AND DEST OPTIONS AND CHANGE VALUES ACCORDINGLY
      xdr.index += Xdr.INT;
      short headerLen = xdr.SHORT();
      packetLen = (short)(headerLen * Xdr.INT + Xdr.INT + cap.length());
      xdr.len = packetLen;
      xdr.PUT(packetLen);
      while (xdr.index < headerLen * Xdr.INT) {
	short optType = (short)(xdr.SHORT() & 0x3FFF);
	short optLen = xdr.SHORT();
	if (optType == DESTOPTTYPE) {
	  xdr.index += Xdr.INT;   // Skip over scheme identifier
	  xdr.PUT(cap.getDst());
	  break;
	} else {
	  xdr.index += (optLen - 1) * Xdr.INT;
	}
      }
      xdr.index = headerLen * Xdr.INT + Xdr.INT;
    }
    else {
      packetLen = (short)(32 + Xdr.INT + cap.length());
      xdr = new Xdr(txbuf, 0, packetLen);
      // Insert Header
      xdr.PUT(ANEPVERSION);
      xdr.PUT((byte)0);  // Forward packet if TypeID unknown
      xdr.PUT(ANTSTYPE);

      // Header with src and dest options = 8 32-bit words
      xdr.PUT((short)8);
      xdr.PUT(packetLen);

      xdr.PUT(DESTOPTTYPE);
      xdr.PUT((short)3);
      xdr.PUT(IPV4SCHEME);
      xdr.PUT(cap.getDst());

      xdr.PUT(SRCOPTTYPE);
      xdr.PUT((short)3);
      xdr.PUT(IPV4SCHEME);
      xdr.PUT(cap.getSrc());

      xdr.PUT(ANTSVERSION);
    }
    cap.xdr = xdr;
    cap.serialize();

    return xdr;
  }

  public Capsule deserialize(Xdr xdr) throws Exception {
    
    if (xdr.BYTE() != ANEPVERSION)
      throw new Exception("Wrong ANEP version");
    boolean discard = (xdr.BYTE() & 0x80) != 0;
    short typeID = xdr.SHORT();
    short headerLen = xdr.SHORT();
    xdr.index += Xdr.SHORT;     // Skip packet length field
    
    if (typeID != ANTSTYPE) {
      if (discard)
	throw new Exception("Unknown type discard");
      else {
	int dest = 0;
	while (xdr.index < headerLen * Xdr.INT) {
	  short flagsAndType = xdr.SHORT();
	  boolean optPrivate = (flagsAndType & 0x8000) != 0;
	  boolean optDiscard = (flagsAndType & 0x4000) != 0;	
	  short optType = (short)(xdr.SHORT() & 0x3FFF);
	  short optLen = xdr.SHORT();
	  if (!optPrivate) {
	    if (optType == DESTOPTTYPE) {
	      int scheme = xdr.INT();
	      if (scheme != IPV4SCHEME)
		throw new Exception("Unknown addressing scheme");
	      else {
		dest = xdr.INT();
	      }
	    } else if (optType == SRCOPTTYPE) {
	      xdr.index += (optLen - 1) * Xdr.INT;
	    } else if (optDiscard) 
	      throw new Exception("Unknown option discard");
	  } else {
	    xdr.index += (optLen - 1) * Xdr.INT;
	  }
	}

	RouteEntry r = client.getRoutes().get(dest);
	if (r != null) {
	  if (xdr.offset != 0) {
	    byte[] bytes = new byte[xdr.len];
	    System.arraycopy(xdr.buf, xdr.offset, bytes, 0, xdr.len);
	    send(bytes, xdr.len, r.addr);
	  } else
	    send(xdr.buf, xdr.len, r.addr);
	  throw new Exception("Unknown type forwarded");
	} else
	  throw new Exception("No route to destination while forwarding");
      }
    }
    
    Method m = codecache.fastLookup(xdr.buf, xdr.offset +
				    headerLen * Xdr.INT + Xdr.INT);

    xdr.index = headerLen * Xdr.INT;

    if (xdr.INT() != ANTSVERSION)
      throw new Exception("Wrong ANTS version");
    
    if (m == null || m.cl == null)
      return new DLBootstrapCapsule(m, xdr, client.getAddress());
    
    if (LOGGING)
      log(L[0], "deserializing "+ m.methodID + " "+ m.cl +
	  " ("+ xdr.length() +" bytes)");
    
    return Capsule.create(m.cl, xdr);
  }
  
  public void attachNode(Node c) {
    if (client != c) {
      client = c;
      clientAddress = c.getAddress();
      codecache = client.capsuleCode;
      client.attachChannel(this);
    }
  }
  
  public final boolean send(Capsule cap, ChannelAddress dst) {
    Xdr xdr = serialize(cap);
    if (xdr.offset != 0) {
      byte[] bytes = new byte[xdr.len];
      System.arraycopy(xdr.buf, xdr.offset, bytes, 0, xdr.len);
      return send(bytes, xdr.len, dst);
    }
    else
      return send(xdr.buf, xdr.len, dst);
  }

  abstract boolean send(byte[] buf, int length, ChannelAddress dst);

  abstract Capsule receive() throws Exception;
  
  /** a one-time start of the channel only */
  public void start() {
    // security 
    if (!started) {
      Thread t = new ChannelThread(this);
      t.start();
      started = true;
    }
  }

  static public Channel New(Manager b, String n, String src) 
    throws Exception
  {
    ChannelAddress c = ChannelAddress.New(src);
    
    if (c instanceof UDPChannelAddress) {
      return new UDPChannel(b, n, (UDPChannelAddress)c);
    } else {
      throw new Exception("unrecognized type of channel" + src);
    }
  }
  
  public Channel(Manager b, String n, ChannelAddress src) {
    this(b, n, src, 32);
  }

  public Channel(Manager b, String n, ChannelAddress src, int q) {
    super(b, n);
    source = src;

    started = false;
  }
}












