/* 
 *  $Id: CodeGroup.java,v 1.7 1997/08/28 18:34:40 djw Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ants;

import java.util.*;
import utils.*;


/**
 * Record collecting information about methods into a related group.
 * Stored in the cache while code is resident or being loaded.
 *
 * @author David Wetherall
 */

class CodeGroup {
  Vector codeIDs;
  Vector meths;
  TypeID groupID;
  Vector groupIDs;
  TypeID protocolID;
  int parts;
  int chunks;
  Vector waiters;
  CodeGroup next;
  CodeGroup prev;
  boolean pin;

  private TypeID computeGroupID() {
    TypeID.sortVector(codeIDs);
    MD5 id = new MD5();
    for (int i = 0; i < codeIDs.size(); i++) {
      TypeID t = (TypeID)(codeIDs.elementAt(i));
      id.update(t.unwrap());
    }
    return TypeID.wrap(id.hash());
  }
 
  void setGroupID() {
    groupID = computeGroupID().intern();
  }
  
  boolean checkGroupID() {
    groupID = groupID.intern();
    return groupID.equals(computeGroupID());
  }

  private TypeID computeProtocolID() {
    TypeID.sortVector(groupIDs);
    MD5 id = new MD5();
    for (int i = 0; i < groupIDs.size(); i++) {
      TypeID t = (TypeID)(groupIDs.elementAt(i));
      id.update(t.unwrap());
    }
    return TypeID.wrap(id.hash());
  }
  
  void setProtocolID() {
    protocolID = computeProtocolID().intern();
  }
  
  void setChunks() throws Exception {
    for (int i = 0; i < meths.size(); i++) {
      Method m = (Method)(meths.elementAt(i));
      chunks += pieces(m.code);
    }

    // start with all parts present
    parts = -1;

    // sanity check limit
    if (chunks > 32)
      throw new Exception("CodeGroup "+ groupID +" too long");
  }

  final static int CHUNK = 1024;

  final short part(int i) {
    return (short)((1<<i) | (-1)<<chunks);
  }

  static int pieces(byte[] code) {
    return (code.length+CHUNK-1) / CHUNK;
  }
  
  final void enqueue(Method meth, Xdr pkt) {
    waiters.addElement(new Waiter(meth, pkt));
  }

  final void enqueue(Waiter waiter) {
    waiters.addElement(waiter);
  }

  final void activate(Node n) {
    for (int i = 0; i < meths.size(); i++) {
      Method m = (Method)meths.elementAt(i);
      m.setCodeID();
      codeIDs.addElement(m.codeID);
    }

    if (!checkGroupID())
      Entity.error("Bad code group identifier");

    setProtocolID();    

    for (int i = 0; i < meths.size(); i++) {
      Method m = (Method)meths.elementAt(i);

      // verify code, group and protocol match method
      if (!m.checkMethodID()) {
	Entity.error("Bad protocol identifier");
      }

      // bootstrap
      m.registerIDs();
    }

    n.loader.loadOnDemand(this);
  }

  final void add(Method m) {
    m.group = this;
    meths.addElement(m);
  }

  CodeGroup() {
    this.meths = new Vector();
    this.codeIDs = new Vector();
    this.groupIDs = new Vector();
    this.waiters = new Vector();
  }

  CodeGroup(TypeID groupID) {
    this();
    this.groupID = groupID;
  }
}


