/* 
 *  $Id: DLResponseCapsule.java,v 1.19 1997/08/31 14:41:02 djw Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ants;


/**
 * Demand Load Response capsule and processing
 *
 * @author David Wetherall
 */


public class DLResponseCapsule extends Capsule {
  final private static byte[] MID = findMID("ants.DLResponseCapsule");
  protected byte[] mid() { return MID; }

  final private static byte[] PID = findPID("ants.DLResponseCapsule");
  protected byte[] pid() { return PID; }

  private TypeID target;
  private TypeID group;
  private TypeID[] groups;
  private int part;
  private int clen;
  private int coff;
  private String _cn;
  private ByteArray code;

  public int length() {
    int s = Xdr.TYPEID + Xdr.TYPEID + Xdr.SHORT + groups.length*Xdr.TYPEID +
      Xdr.INT + Xdr.INT + Xdr.INT + Xdr.STRING(_cn) + Xdr.BYTEARRAY(code);
    return super.length() + s;
  }
  
  public Xdr serialize() {
    Xdr xdr = super.serialize();
    xdr.PUT(target);
    xdr.PUT(group);
    xdr.PUT((short)groups.length);
    for (int n = 0; n < groups.length; n++)
      xdr.PUT(groups[n]);
    xdr.PUT(part);
    xdr.PUT(clen);
    xdr.PUT(coff);
    xdr.PUT(_cn);
    xdr.PUT(code);

    return xdr;
  }

  public Xdr deserialize() {
    Xdr xdr = super.deserialize();
    target = xdr.TYPEID();
    group = xdr.TYPEID();
    int l = xdr.SHORT();
    groups = new TypeID[l];
    for (int n = 0; n < l; n++) {
      groups[n] = xdr.TYPEID();
    }
    part = xdr.INT();
    clen = xdr.INT();
    coff = xdr.INT();
    _cn = xdr.STRING();
    code = xdr.BYTEARRAY();
    
    return xdr;
  }
  
  public boolean evaluate(Node n) {
    if (Entity.LOGGING)
      n.log(Entity.L[0], "deserialize: received DLResponse "+
	    target +" from "+ NodeAddress.toString(getSrc()));
    
    Method m = n.capsuleCode.lookup(target);
    if (m == null) {
      m = new Method(target);
      n.capsuleCode.enter(m);
    }

    CodeGroup g = n.capsuleCode.retrieve(group);
    if (g == null) {
      g = new CodeGroup(group);
      n.capsuleCode.set(g);
    } 

    if (m.group == null) m.join(g);
    if (g.groupIDs.size() ==  0) {
      for (int i = 0; i < groups.length; i++) {
	g.groupIDs.addElement(groups[i]);
      }
    }

    if ((g.parts | part) != g.parts) {
      if (m.code == null) {
	
	if (Entity.LOGGING)
	  n.log(Entity.L[0], "DLResponse: allocating "+ clen +" for "+ _cn);
	
	m.code = new byte[clen];
	m._cn = _cn;
	g.chunks += g.pieces(m.code);
      }
      
      if (Entity.LOGGING)
	n.log(Entity.L[0], "DLResponse: adding to "+ _cn + " (off= "+
	      coff +", part= "+ Integer.toHexString(part) + ", len= " +
	      code.length() + ")");
      
      code.getBytes(0, m.code, coff, code.length());
      g.parts |= part;
      
      if (g.parts == -1) {
	
	if (Entity.LOGGING)
	  n.log(Entity.L[0], "DLResponse: completed with "+ _cn);

	g.activate(n);

	while (!g.waiters.isEmpty()) {
	  Waiter w = (Waiter)g.waiters.firstElement();
	  Capsule cap = Capsule.create(w.meth.cl, w.waiter);

	  try {

	    // make it sequential
	    synchronized(n) {

	      // wrap context since we are a recursive call
	      Capsule saved = n.context;
	      try {

		n.context = cap;
		cap.evaluate(n);

	      } finally {
		n.context = saved;
	      }
	    }

	  } catch (Exception e) {
	    if (Entity.LOGGING)
	      n.log(Entity.L[0], "terminated "+ cap +" with "+ e);
	  }

	  g.waiters.removeElementAt(0);
	}  
      }
    } else if (Entity.LOGGING)
      n.log(Entity.L[0], "deserialize: unneeded DLResponse "+
	    target +" from "+ NodeAddress.toString(getSrc()));
    
    return true;
  }
  

  DLResponseCapsule() { }

  DLResponseCapsule(Method m, int idx, int pos, int len) {
    target = m.methodID;
    group = m.group.groupID;
    groups = new TypeID[m.group.groupIDs.size()];
    for (int i = 0; i < m.group.groupIDs.size(); i++) {
      groups[i] = (TypeID)(m.group.groupIDs.elementAt(i));
    }
    part = m.group.part(idx);
    clen = m.code.length;
    coff = pos;
    _cn = m._cn;

    code = new ByteArray(m.code, coff, len);
  }
}


