/* 
 *  $Id: Entity.java,v 1.10 1997/08/30 15:29:06 djw Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ants;

import java.awt.*;
import java.util.Date;
import utils.*;


/**
 * Shell of the entities in that are combined to produce a running 
 * node configuration.  Must inherit from Frame to provide GUI potential
 * the way the AWT is set up -- but this is somewhat of a problem
 * since we squander Frames when we don't want them and use up our
 * single inheritance slot.
 *
 * @author David Wetherall
 */


abstract class Entity
  extends Frame
{
  final public static String[] defaults = { "-log", "0" };

  /**
   * Switch to control overall logging.  If false, we expect log calls
   * to be removed at compile-time by dead-code elimination and inlining
   */

  final static boolean LOGGING = false;
  final static public int[] L = { 1, 2, 4, 8, 16, 32, 64, 128 };
  final static int LOG_LEVELS = L.length;

  /** Vector of bit flags for configurable logging. Starts disabled. */
  protected int logLevel = 0;

  /** Only settable if we are logging */
  public void setLogLevel(int l) { if (LOGGING) logLevel = l; }
  public int getLogLevel() { return logLevel; }
  
  /** Report an event to our manager if relevant. */
  abstract protected void log(int level, String msg);
  
  public String getName() { return getTitle(); }
  public void setName(String n) { setTitle(n); }


  private static String stamp() {
    Date d = new Date();
    return d.toString().substring(4, 20);
  }

  public static void error(String msg) {
    System.err.println(stamp() + "TERMINATING WITH ERROR-- " + msg);
    new Throwable(msg).printStackTrace();
    System.exit(1);
  }

  public static void warn(String msg) {
    System.err.println(stamp() + "WARNING-- " + msg);
  }

  public static void info(String msg) {
    System.err.println(stamp() + msg);
  }

  final static boolean ASSERTING = true;

  protected static void assert(boolean ok, String msg) {
    if (ASSERTING && !ok) error(msg);
  }

  protected void finalize() throws Throwable {
    super.finalize();
    // a chance to dump statistics, profile info, etc.
  }
  
  /** Catch our window being closed as a hint bail */
  public boolean handleEvent(Event evt) {
    if (evt.id == Event.WINDOW_DESTROY) {
      dispose();
      System.exit(0);
      return true;
    } else 
      return super.handleEvent(evt);
  }

  //
  // usage is waiting for reflection in 1.1 ...
  //
  //  public String usage()
  //    throws Exception
  //  {
  //  Class top = getClass();
  //  Class base = Class.forName("ants.Entity");
  //  KeyArgs k = new KeyArgs(defaults);
  //
  //    String us = top.getName() + ": ";
  //    while (top != base) {
  //      try {
  //	if (((top)this).defaults != null)
  //	  k.merge(((top)this).defaults);
  //      } catch (NoSuchFieldError e) {
  //      }
  //      top = top.getSuperclass();
  //    }
  //    return us + k.toString();
  //  }
  
  public void setArgs(KeyArgs k) 
    throws Exception 
  {
    k.merge(defaults);

    for (int i = 0; i < k.length(); i++) {
      if (k.key(i).equals("-log")) {
	setLogLevel(Integer.parseInt(k.arg(i)));
	k.strike(i);
      }
    }

    k.compact();
    if (k.length() > 0) throw new Exception("leftover keyargs: " + k);
  }
  
  public void setArgs(String[] ka) 
    throws Exception 
  {
    setArgs(ka, 0);
  }
  
  public void setArgs(String[] ka, int sk)
    throws Exception
  { 
    setArgs(new KeyArgs(ka, sk));
  }

  public void start()
    throws Exception 
  {
  }

  /** No-argument constructor if called through newInstance */
  public Entity() { super(); }

  /** Normal construction gives entity a name */
  public Entity(String t) { super(); setName(t); }
}


