/* 
 *  $Id: Method.java,v 1.9 1997/08/28 18:34:42 djw Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */

package ants;

import java.util.*;
import utils.*;


/** 
 * This class (along with CodeGroups) is used to represent all the
 * information about a type of capsule in the code cache.  It is used
 * only within ants, from Protocol during registration, Channel during
 * capsule execution dispatch, and DL* during code loading.
 *
 * @author David Wetherall
 */

final class Method {

  // the parts of method have package access since it may be in many
  // different states of completion during code loading and protocol
  // registration

  byte[] code;
  String _cn;
  Class cl;
  TypeID codeID;
  int hash;
  TypeID methodID;
  CodeGroup group;
  Vector waiters;
  Method next;

  // until we have a better way to initialize the 
  // methodid of a capsule, we keep this side table
  // based on local names
  
  /** Local class name to methodID table for bootstrapping */
  private static Hashtable initializers = new Hashtable();

  /** Entry point used for protocol definition and loading */
  void registerIDs() {
    TypeID[] ids = new TypeID[3];
    ids[0] = methodID;
    ids[1] = group.groupID;
    ids[2] = group.protocolID;

    // check that repeat registrations match
    TypeID[] t = (TypeID[])initializers.get(_cn);
    if (t != null) {
      if (t[0] != ids[0] || t[1] != ids[1] || t[2] != ids[2]) 
	Entity.error("Different registration for " + _cn);
    }
    
    initializers.put(_cn, ids);
  }

  /** Access point to initialze capsule methodIDs */
  static byte[] findMID(String name) {
    TypeID[] t = (TypeID[])initializers.get(name);
    if (t == null) Entity.error("No registration for "+ name);

    // we also use this opportunity to verify the load path
    DLClassLoader.verifyOrigin(name);

    return t[0].unwrap();
  }
  
  /** Access point to initialze capsule groupIDs */
  static byte[] findGID(String name) {
    TypeID[] t = (TypeID[])initializers.get(name);
    if (t == null) Entity.error("No registration for "+ name);
    return t[1].unwrap();
  }
  
  /** Access point to initialze capsule protocolIDs */
  static byte[] findPID(String name) {
    TypeID[] t = (TypeID[])initializers.get(name);
    if (t == null) Entity.error("No registration for "+ name);
    return t[2].unwrap();
  }
  
  private TypeID computeMethodID() {
    MD5 id = new MD5();

    // code, group, protocol is the canonical order
    id.update(codeID.unwrap());
    id.update(group.groupID.unwrap());
    id.update(group.protocolID.unwrap());
    return TypeID.wrap(id.hash());
  }

  void setMethodID() {
    methodID = computeMethodID().intern();
  }

  void setCodeID() { 
    codeID = TypeID.wrap(MD5.hash(code));
  }

  boolean checkMethodID() {
    methodID = methodID.intern();
    return methodID.equals(computeMethodID());
  }

  
  void enqueue(Xdr pkt) {
    waiters.addElement(new Waiter(this, pkt));
  }

  void join(CodeGroup g) {
    group = g;
    g.add(this);
    for (int i = 0; i < waiters.size(); i++) {
      g.enqueue((Waiter)waiters.elementAt(i));
    }
    waiters = null;
  }
  
  Method(TypeID methodID) {
    this.methodID = methodID;
    this.waiters = new Vector();
  }

  /** Construct and initialize from the local class name */
  Method(String name) throws Exception {

    // until we control the local name under which 
    // a method is loaded, we need to save this info
    _cn = name;
    
    code = ClassCode.get(name);
    setCodeID();
    waiters = new Vector();
  }
}
