/* 
 *  $Id: Node.java,v 1.41 1997/08/31 14:50:48 djw Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ants;

import java.net.*;
import java.util.*;

import ants.wrapper.*;
import utils.*;


/**
 * node of an active network
 *
 * @author David Wetherall
 */

public class Node
  extends ManagedObject
  implements Runnable
{
  private boolean running;
  private String routefile;
  Capsule context; // accessed by ChannelThreads to set context

  private NodeSecurityManager secure;

  /** Secured loader, accessed directly by codegroups within ants */
  DLClassLoader loader;

  private int address;
  
  public int getAddress() { 
    return address;
  }

  private RouteTable routes;
  
  public RouteTable getRoutes() {
    checkCapsule();
    return routes;
  }

  private Hashtable clients;

  public Hashtable getApplications() {
    checkCapsule();
    return clients;
  }

  private Channel link;

  public Channel getChannels() {
    checkCapsule();
    return link;
  }

  private Extension[] extensions;

  public Extension[] getExtensions() {
    checkCapsule();
    return extensions;
  }

  private AutoNodeCache cache;

  public AutoNodeCache getCache() { 
    return cache;
  }
  
  CodeCache capsuleCode;
  
  public void setArgs(KeyArgs k) 
    throws Exception
  {
    k.merge(defaults);

    for (int i = 0; i < k.length(); i++) {
      if (k.key(i).equals("-routes")) {
	routefile = k.arg(i);
	k.strike(i);
      }
    }
    
    super.setArgs(k);
  }

  public void start() 
    throws Exception 
  {
    Entity.info("Node " + NodeAddress.toString(address) + " starts");

    if (routefile != null) {
      Entity.info("Reading routes from " + routefile);
      routes.read(routefile);
    }

    // export the node environment to capsule classes
    loader.exportClass("ants.Node");
    loader.exportClass("ants.Capsule");
    loader.exportClass("ants.Xdr");
    loader.exportClass("ants.NodeAddress");
    loader.exportClass("ants.ByteArray");

    // install the system protocols
    register(new DLProtocol());
    register(new DataProtocol());

    // install ourselves as a security mediator
    secure.addMediator(this.getClass());

    // start any node extensions
    for (int i = 0; i < extensions.length; i++) {
      extensions[i].start();
    }

    // now the node is complete
    running = true;

    // start application and network
    Enumeration e = clients.elements();
    while (e.hasMoreElements()) 
      ((Application)e.nextElement()).start();

    if (link != null) link.start();

    super.start();
  }

  /** Register a protocol for use at this node */
  public void register(Protocol p) 
    throws Exception
  {
    for (int i = 0; i < p.groups.size(); i++) {
      CodeGroup g = (CodeGroup)p.groups.elementAt(i);

      loader.loadForRegister(g);

      if (capsuleCode.retrieve(g.groupID) == null) {
	g.pin = true;
	capsuleCode.set(g);
      }
      
      for (int j = 0; j < g.meths.size(); j++) {
	Method m = (Method)g.meths.elementAt(j);

	if (capsuleCode.lookup(m.methodID) == null) {
	  capsuleCode.enter(m);
	}
      }
    }
  }

  /** Unregister a protocol previously registered at this node */
  public void unregister(Protocol p) {
    // sorry, just an api call for now
  }

  public void attachApplication(Application c, int port) throws Exception {
    Integer key = new Integer(port);

    if (clients.containsKey(key))
      throw new Exception("Port " + port + " is busy.");
    
    clients.put(key, c);
  }
  
  public void attachChannel(Channel l) {
    if (link != l) {
      link = l;
      link.attachNode(this);
    }
  }

  /** For capsules to quickly retrieve known extensions */
  public Extension getExtension(int index) {
    return extensions[index];
  }

  public void attachExtension(Extension e) {
    int index = Extension.registerNamedExtension(e.getName());

    if (extensions.length < index+1) {
      Extension[] ne = new Extension[index+1];
      System.arraycopy(extensions, 0, ne, 0, extensions.length);
      extensions = ne;
    }

    if (extensions[index] != e){
      extensions[index] = e;
      e.attachNode(this);
    } 
  }

  /** Export a class for use by capsules */
  public void exportClass(String name) throws ClassNotFoundException {
    secure.checkCapsule();
    loader.exportClass(name);
  }

  /** Security check to see if a capsule directs the invocation */
  public void checkCapsule() {
    
    // start the check at our caller's site
    secure.checkCapsule(1);
  }

  /** Add a class to the list of security check mediators */
  public void addMediator(Class cl) {
    secure.checkCapsule();
    secure.addMediator(cl);
  }

  public boolean routeForNode(Capsule c, int dst) {
    RouteEntry r = routes.get(dst);
    
    // Log("dst = " + dst);
    
    if (r != null) {
      // Log("routing to " + r.nxt);
      return link.send(c, r.addr);
    } else
      return false;
  }
  
  public boolean sendToNeighbors(Capsule c) {//ulana
    int i;
    int ok = 0;
    RouteEntry r;

    for (i = 0; i < routes.neighbor_list.length; i++) {
       r = routes.neighbor_list[i];
       if (r != null)
	  if (link.send(c, r.addr) == true) ok++;
    }
    if (ok < routes.neighbor_list.length)
      return false;
    else
      return true;
  }

  public boolean deliverToApp(Capsule c, int d) {
    c.saveBuffer = true;
    ((Application)clients.get(new Integer(d))).receive(c);
    return true;
  }


  public long time() {
    return System.currentTimeMillis();
  }

  public void log(String msg) {
    Entity.info(this.context.getClass().getName() + ": " + msg);
  }

  /** Called by Application base class to send a capsule */ 
  synchronized boolean send(Capsule c) {
    this.context = c;
    return c.evaluate(this);
  }

  /** For use by node components and applications to delay execution */
  public void sleep(long millis) {
    checkCapsule();

    try {
      Thread.sleep(millis);
    } catch (InterruptedException e) {
      ; // who cares
    }
  }
  
  public void run() {
    checkCapsule();
  }


 /** constructs a node with given address */
  public Node(Manager b, String n, String a)
    throws Exception
  {
    super(b, n);

    // first, check system integrity
    secure = new NodeSecurityManager();
    secure.checkCapsule();

    running = false;
    address = NodeAddress.fromString(a);
    routes = new RouteTable(address);
    cache = new AutoNodeCache(this, 1024, 600);
    capsuleCode = new CodeCache(256);
    loader = new DLClassLoader();
    extensions = new Extension[0];
    clients = new Hashtable();
  }

}
